"""
A script to collect a batch of human demonstrations that can be used
to generate a learning curriculum (see `demo_learning_curriculum.py`).

The demonstrations can be played back using the `playback_demonstrations_from_pkl.py`
script.
"""

import argparse
from copy import deepcopy
import datetime
import json
import os
import time
from glob import glob
import sys
import tarfile
import subprocess
import h5py
import imageio
import mujoco
import numpy as np
import platform
import robosuite

# from robosuite import load_controller_config
from robosuite.controllers import load_composite_controller_config
from robosuite.wrappers import DataCollectionWrapper, VisualizationWrapper
from termcolor import colored

import robocasa
import robocasa.macros as macros
from robocasa.models.fixtures import FixtureType
from robocasa.utils.robomimic.robomimic_dataset_utils import convert_to_robomimic_format
from robocasa_upload.UploadGenerator import UploadGenerator

DELTA_INPUT_DEVICES = ["spacemouse"]

def collect_human_trajectory(
    env,
    device,
    arm,
    env_configuration,
    mirror_actions,
    render=True,
    max_fr=None,
    print_info=True,
):
    """
    Use the device (keyboard or SpaceNav 3D mouse) to collect a demonstration.
    The rollout trajectory is saved to files in npz format.
    Modify the DataCollectionWrapper wrapper to add new fields or change data formats.

    Args:
        env (MujocoEnv): environment to control
        device (Device): to receive controls from the device
        arms (str): which arm to control (eg bimanual) 'right' or 'left'
        env_configuration (str): specified environment configuration
    """

    env.reset()

    ep_meta = env.get_ep_meta()
    # print(json.dumps(ep_meta, indent=4))
    lang = ep_meta.get("lang", None)
    if print_info and lang is not None:
        print(colored(f"Instruction: {lang}", "green"))

    # degugging: code block here to quickly test and close env
    # env.close()
    # return None, True

    if render:
        # ID = 2 always corresponds to agentview
        env.render()

    task_completion_hold_count = (
        -1
    )  # counter to collect 10 timesteps after reaching goal
    device.start_control()

    nonzero_ac_seen = False

    # Keep track of prev gripper actions when using since they are position-based and must be maintained when arms switched
    all_prev_gripper_actions = [
        {
            f"{robot_arm}_gripper": np.repeat([0], robot.gripper[robot_arm].dof)
            for robot_arm in robot.arms
            if robot.gripper[robot_arm].dof > 0
        }
        for robot in env.robots
    ]

    zero_action = np.zeros(env.action_dim)
    for _ in range(1):
        # do a dummy step thru base env to initalize things, but don't record the step
        if isinstance(env, DataCollectionWrapper):
            env.env.step(zero_action)
        else:
            env.step(zero_action)

    discard_traj = False
    reset_status = False

    # Loop until we get a reset from the input or the task completes
    while True:
        start = time.time()

        # Set active robot
        active_robot = env.robots[device.active_robot]
        active_arm = device.active_arm

        # Get the newest action
        input_ac_dict = device.input2action(mirror_actions=mirror_actions)

        # If action is none, then this a reset so we should break
        if input_ac_dict is None:
            discard_traj = True
            reset_status = True
            break

        action_dict = deepcopy(input_ac_dict)

        # set arm actions
        for arm in active_robot.arms:
            controller_input_type = active_robot.part_controllers[arm].input_type
            if controller_input_type == "delta":
                action_dict[arm] = input_ac_dict[f"{arm}_delta"]
            elif controller_input_type == "absolute":
                action_dict[arm] = input_ac_dict[f"{arm}_abs"]
            else:
                raise ValueError
        # print(action_dict)
        if device.is_empty_input(action_dict):
            if not nonzero_ac_seen:
                if render:
                    env.render()
                continue
        else:
            nonzero_ac_seen = True

        # Maintain gripper state for each robot but only update the active robot with action
        env_action = [
            robot.create_action_vector(all_prev_gripper_actions[i])
            for i, robot in enumerate(env.robots)
        ]
        env_action[device.active_robot] = active_robot.create_action_vector(action_dict)
        env_action = np.concatenate(env_action)

        # Run environment step
        obs, _, _, _ = env.step(env_action)
        if render:
            env.render()

        # Also break if we complete the task
        if task_completion_hold_count == 0:
            break

        # state machine to check for having a success for 15 consecutive timesteps
        if env._check_success():
            if task_completion_hold_count > 0:
                task_completion_hold_count -= 1  # latched state, decrement count
            else:
                task_completion_hold_count = 15  # reset count on first success timestep
        else:
            task_completion_hold_count = -1  # null the counter if there's no success

        # limit frame rate if necessary
        if max_fr is not None:
            elapsed = time.time() - start
            diff = 1 / max_fr - elapsed
            if diff > 0:
                time.sleep(diff)

        # with open("/home/soroushn/tmp/model.xml", "w") as f:
        #     f.write(env.model.get_xml())
        # exit()

    if nonzero_ac_seen and hasattr(env, "ep_directory"):
        ep_directory = env.ep_directory
    else:
        ep_directory = None

    # cleanup for end of data collection episodes
    env.close()

    return ep_directory, discard_traj, reset_status


def gather_demonstrations_as_hdf5(
    directory,
    out_dir,
    env_info,
    successful_episodes=None,
    verbose=False,
    out_name="demo.hdf5",
):
    """
    Gathers the demonstrations saved in @directory into a
    single hdf5 file.
    The strucure of the hdf5 file is as follows.
    data (group)
        date (attribute) - date of collection
        time (attribute) - time of collection
        repository_version (attribute) - repository version used during collection
        env (attribute) - environment name on which demos were collected
        demo1 (group) - every demonstration has a group
            model_file (attribute) - model xml string for demonstration
            states (dataset) - flattened mujoco states
            actions (dataset) - actions applied during demonstration
        demo2 (group)
        ...
    Args:
        directory (str): Path to the directory containing raw demonstrations.
        out_dir (str): Path to where to store the hdf5 file.
        env_info (str): JSON-encoded string containing environment information,
            including controller and robot info
    """

    hdf5_path = os.path.join(out_dir, out_name)
    if verbose:
        print("Saving hdf5 to", hdf5_path)
    f = h5py.File(hdf5_path, "w")

    # store some metadata in the attributes of one group
    grp = f.create_group("data")

    num_eps = 0
    env_name = None  # will get populated at some point

    for ep_directory in os.listdir(directory):
        # print("Processing {} ...".format(ep_directory))
        if (successful_episodes is not None) and (
            ep_directory not in successful_episodes
        ):
            # print("\tExcluding this episode!")
            continue

        state_paths = os.path.join(directory, ep_directory, "state_*.npz")
        states = []
        actions = []
        actions_abs = []
        # success = False

        for state_file in sorted(glob(state_paths)):
            dic = np.load(state_file, allow_pickle=True)
            env_name = str(dic["env"])

            states.extend(dic["states"])
            for ai in dic["action_infos"]:
                actions.append(ai["actions"])
                if "actions_abs" in ai:
                    actions_abs.append(ai["actions_abs"])
            # success = success or dic["successful"]

        if len(states) == 0:
            continue

        # # Add only the successful demonstration to dataset
        # if success:

        # print("Demonstration is successful and has been saved")
        # Delete the last state. This is because when the DataCollector wrapper
        # recorded the states and actions, the states were recorded AFTER playing that action,
        # so we end up with an extra state at the end.
        del states[-1]
        assert len(states) == len(actions)

        num_eps += 1
        ep_data_grp = grp.create_group("demo_{}".format(num_eps))

        # store model xml as an attribute
        xml_path = os.path.join(directory, ep_directory, "model.xml")
        with open(xml_path, "r") as f:
            xml_str = f.read()
        ep_data_grp.attrs["model_file"] = xml_str

        # store ep meta as an attribute
        ep_meta_path = os.path.join(directory, ep_directory, "ep_meta.json")
        if os.path.exists(ep_meta_path):
            with open(ep_meta_path, "r") as f:
                ep_meta = f.read()
            ep_data_grp.attrs["ep_meta"] = ep_meta

        # write datasets for states and actions
        ep_data_grp.create_dataset("states", data=np.array(states))
        ep_data_grp.create_dataset("actions", data=np.array(actions))
        if len(actions_abs) > 0:
            print(np.array(actions_abs).shape)
            ep_data_grp.create_dataset("actions_abs", data=np.array(actions_abs))

        # else:
        #     pass
        #     # print("Demonstration is unsuccessful and has NOT been saved")

    if verbose:
        print("{} successful demos so far".format(num_eps))

    if num_eps == 0:
        f.close()
        return

    # write dataset attributes (metadata)
    now = datetime.datetime.now()
    grp.attrs["date"] = "{}-{}-{}".format(now.month, now.day, now.year)
    grp.attrs["time"] = "{}:{}:{}".format(now.hour, now.minute, now.second)
    grp.attrs["robocasa_version"] = robocasa.__version__
    grp.attrs["robosuite_version"] = robosuite.__version__
    grp.attrs["mujoco_version"] = mujoco.__version__
    grp.attrs["env"] = env_name
    grp.attrs["env_info"] = env_info

    f.close()

    return hdf5_path

if __name__ == "__main__":
    # Arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--directory",
        type=str,
        default=os.path.join(robocasa.models.assets_root, "demonstrations_private"),
    )
    parser.add_argument("--environment", type=str, default="Kitchen")
    parser.add_argument(
        "--robots",
        nargs="+",
        type=str,
        default="PandaOmron",
        help="Which robot(s) to use in the env",
    )
    parser.add_argument(
        "--config",
        type=str,
        default="single-arm-opposed",
        help="Specified environment configuration if necessary",
    )
    parser.add_argument(
        "--arm",
        type=str,
        default="right",
        help="Which arm to control (eg bimanual) 'right' or 'left'",
    )
    parser.add_argument(
        "--obj_groups",
        type=str,
        nargs="+",
        default=None,
        help="In kitchen environments, either the name of a group to sample object from or path to an .xml file",
    )

    parser.add_argument(
        "--camera",
        type=str,
        default=None,
        help="Which camera to use for collecting demos",
    )
    parser.add_argument(
        "--controller",
        type=str,
        default=None,
        help="Choice of controller. Can be, eg. 'NONE' or 'WHOLE_BODY_IK', etc. Or path to controller json file",
    )
    parser.add_argument(
        "--device",
        type=str,
        default="spacemouse",
        choices=["keyboard", "spacemouse", "vr-controller", "vr-hand"],
    )
    parser.add_argument(
        "--pos-sensitivity",
        type=float,
        default=4.0,
        help="How much to scale position user inputs",
    )
    parser.add_argument(
        "--rot-sensitivity",
        type=float,
        default=4.0,
        help="How much to scale rotation user inputs",
    )
    parser.add_argument(
        "--continuous-pressing",
        type=bool,
        default=False,
        help="Whether to use continuous pressing for spacemouse",
    )

    parser.add_argument("--debug", action="store_true")
    parser.add_argument(
        "--renderer", type=str, default="mjviewer", choices=["mjviewer", "mujoco"]
    )
    parser.add_argument(
        "--max_fr", default=30, type=int, help="If specified, limit the frame rate"
    )

    parser.add_argument("--layout", type=int, nargs="+", default=-1)
    parser.add_argument(
        "--style", type=int, nargs="+", default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 11]
    )
    parser.add_argument("--generative_textures", action="store_true")
    args = parser.parse_args()

    # assert args.layout == -1, colored("Currently, layout does not allow fixed", "red")

    # Get controller config
    # controller_config = load_controller_config(default_controller=args.controller)
    controller_config = load_composite_controller_config(
        controller=args.controller,
        robot=args.robots if isinstance(args.robots, str) else args.robots[0],
    )

    if args.device in DELTA_INPUT_DEVICES:
        controller_config["body_parts"]["right"]["input_type"] = "delta"
        print(colored(f"Using {args.device}. Set delta input type for arms", "red"))

    if controller_config["type"] == "WHOLE_BODY_MINK_IK":
        # mink-speicific import. requires installing mink
        from robosuite.examples.third_party_controller.mink_controller import (
            WholeBodyMinkIK,
        )

    env_name = args.environment

    # Create argument configuration
    config = {
        "env_name": env_name,
        "robots": args.robots,
        "controller_configs": controller_config,
    }

    if args.generative_textures is True:
        config["generative_textures"] = "100p"

    # Check if we're using a multi-armed environment and use env_configuration argument if so
    if "TwoArm" in env_name:
        config["env_configuration"] = args.config

    # Mirror actions if using a kitchen environment
    if env_name in ["Lift"]:  # add other non-kitchen tasks here
        if args.obj_groups is not None:
            print(
                "Specifying 'obj_groups' in non-kitchen environment does not have an effect."
            )
        mirror_actions = False
        if args.camera is None:
            args.camera = "agentview"
        # special logic: "free" camera corresponds to Null camera
        elif args.camera == "free":
            args.camera = None
    else:
        mirror_actions = True
        if args.layout is not None:
            config["layout_ids"] = args.layout
        if args.style is not None:
            config["style_ids"] = args.style
        ### update config for kitchen envs ###
        if args.obj_groups is not None:
            config.update({"obj_groups": args.obj_groups})
        if args.camera is None:
            args.camera = "robot0_frontview"
        # special logic: "free" camera corresponds to Null camera
        elif args.camera == "free":
            args.camera = None

        config["translucent_robot"] = True

        # by default use obj instance split A
        config["obj_instance_split"] = "A"
        # config["obj_instance_split"] = None
        # config["obj_registries"] = ("aigen",)

    if platform.system() == "Windows":
        os.environ["MUJOCO_GL"] = "wgl"
    elif platform.system() == "Darwin":
        os.environ["MUJOCO_GL"] = "cgl"
    elif platform.system() == "Linux":
        os.environ["MUJOCO_GL"] = "egl"
    
    # Create environment
    env = robosuite.make(
        **config,
        has_renderer=True,
        has_offscreen_renderer=False,
        render_camera=args.camera,
        ignore_done=True,
        use_camera_obs=False,
        control_freq=20,
        renderer=args.renderer,
    )
    origin_layout_and_style_ids = env.layout_and_style_ids

    # Wrap this with visualization wrapper
    env = VisualizationWrapper(env)

    # Grab reference to controller config and convert it to json-encoded string
    env_info = json.dumps(config)

    t_now = time.time()
    time_str = datetime.datetime.fromtimestamp(t_now).strftime("%Y-%m-%d-%H-%M-%S")
    time_str = f"{time_str}_{env_name}"

    if not args.debug:
        # make a new timestamped directory
        demo_dir = os.path.join(args.directory, time_str)
        os.makedirs(demo_dir)

    if not args.debug:
        # wrap the environment with data collection wrapper
        all_eps_directory = os.path.join(demo_dir, "episodes")
        env = DataCollectionWrapper(env, all_eps_directory)

    # initialize device
    if args.device == "keyboard":
        from robosuite.devices import Keyboard

        device = Keyboard(
            env=env,
            pos_sensitivity=args.pos_sensitivity,
            rot_sensitivity=args.rot_sensitivity,
        )
    elif args.device == "spacemouse":
        from robosuite.devices import SpaceMouse

        device = SpaceMouse(
            env=env,
            pos_sensitivity=args.pos_sensitivity,
            rot_sensitivity=args.rot_sensitivity,
            vendor_id=macros.SPACEMOUSE_VENDOR_ID,
            product_id=macros.SPACEMOUSE_PRODUCT_ID,
            is_continuous_pressing=args.continuous_pressing,
        )
    elif args.device.startswith("vr"):
        if args.device == "vr-controller":
            from robosuite.devices.vr import VRController as VR
        elif args.device == "vr-hand":
            from robosuite.devices.vr import VRHand as VR
        else:
            raise ValueError(f"Invalid VR device: {args.device}")
        from multiprocessing import Process, shared_memory
        image_size = (720, 1280)
        shm = shared_memory.SharedMemory(
            create=True,
            size=image_size[0] * image_size[1] * 3 * np.uint8().itemsize,
        )
        device = VR(
            env=env,
            img_shape=image_size,
            shm_name=shm.name,
            robot_mode="arm",
        )
    else:
        raise ValueError

    # 在本进程内初始化LogGenerator
    os.environ['PYTHONOPTIMIZE'] = '1'  # 启用Python优化
    upload_gen = UploadGenerator(demo_dir, env_name, env_info)

    successful_episodes = []

    # collect demonstrations
    try:
        while True:
            print()
            ep_directory, discard_traj, reset_status = collect_human_trajectory(
                env,
                device,
                args.arm,
                args.config,
                mirror_actions,
                render=(args.renderer != "mjviewer"),
                max_fr=args.max_fr,
            )

            ep_layout_id = None
            if os.path.exists(os.path.join(ep_directory, "ep_meta.json")):
                with open(os.path.join(ep_directory, "ep_meta.json"), "r") as f:
                    ep_layout_id = json.load(f)["layout_id"]

            # if reset, keep layout
            if reset_status and ep_layout_id is not None:
                # DataCollectionWrapper -> VisualizationWrapper -> MujocoEnv
                env.env.env.layout_and_style_ids = [pair for pair in origin_layout_and_style_ids if pair[0] == ep_layout_id]
            else: env.env.env.layout_and_style_ids = origin_layout_and_style_ids

            if ep_directory is not None:
                # save whether the episode is successful or not
                with open(os.path.join(ep_directory, "ep_stats.json"), "w") as file:
                    json.dump({"success": not discard_traj}, file)

                with open(os.path.join(ep_directory, "env_info.json"), "w") as file:
                    json.dump(env_info, file)

                if not discard_traj:
                    successful_episodes.append(os.path.split(ep_directory)[-1])

                gather_demonstrations_as_hdf5(
                    all_eps_directory,
                    ep_directory,
                    env_info,
                    successful_episodes=[os.path.split(ep_directory)[-1]],
                    out_name="ep_demo.hdf5",
                )

            print("Episode success:", not discard_traj)

            if not args.debug:
                hdf5_path = gather_demonstrations_as_hdf5(
                    all_eps_directory,
                    demo_dir,
                    env_info,
                    successful_episodes=successful_episodes,
                    verbose=True,
                )
                if hdf5_path is not None:
                    convert_to_robomimic_format(hdf5_path)

            if upload_gen is not None and ep_layout_id is not None:
                user_input = input(colored("Upload this episode? (y/n): ", "cyan"))
                while True:
                    if user_input.lower() == 'n':
                        break
                    elif user_input.lower() == 'y':
                        metadata = json.dumps({'robot_name': args.robots if isinstance(args.robots, str) else args.robots[0], 'layout_id': ep_layout_id, 'device_name': args.device, 'system': platform.system(), 'is_success': not discard_traj})
                        upload_gen.upload(os.path.split(ep_directory)[-1], env, metadata)
                        break
                    else:
                        print(colored("Invalid input, please input y/n:", "red"))
                        user_input = input()

    except KeyboardInterrupt:
        print("\nKeyboard interrupt detected. Performing cleanup...")
        if not args.debug:
            hdf5_path = gather_demonstrations_as_hdf5(
                all_eps_directory,
                demo_dir,
                env_info,
                successful_episodes=successful_episodes,
                verbose=True,
            )
            if hdf5_path is not None:
                convert_to_robomimic_format(hdf5_path)
                print(colored(f"\nDataset saved: {hdf5_path}", "green"))
    finally:
        if args.device.startswith("vr"):
            device.close()
            shm.unlink()
            shm.close()
        env.close()
    sys.exit(0)
